
from 木兰.生成.python import 代码生成器 as CodegenNow
from 木兰.分析器.词法分析器 import 分词器 as tokenizer
from 木兰.分析器.语法分析器 import 语法分析器
from 木兰.分析器.错误 import 语法错误

from os import walk
from os.path import join

from typing import Dict
from json import dump, load as json_load
from subprocess import Popen, PIPE
from unittest import TestCase


木兰原始文件执行命令 = 'ulang-0.2.2.exe --dump-python'

原始木兰无法运行 = (
    'signature.ul'
)


def 生成测试用例json文件():
    源码表 = {}

    for 当前目录, 各文件夹, 各文件 in walk('测试/unittest/源码生成/木兰到Python', ):
        for 文件名 in 各文件:
            if 文件名[-3:] == '.ul' and '待解决' not in 文件名 and  \
                    文件名 not in 原始木兰无法运行:
                p = join(当前目录, 文件名)
                print('加载文件 %s...' % p)
                with open(p) as f:
                    cont = f.read()
                    源码表[p] = cont
    with  open('测试/unittest/测试源码表.json', 'w') as f:
        dump(源码表, f)


def 读取测试用例json() -> Dict[str, str]:
    with open('测试/unittest/测试源码表.json', 'r') as f:
        return json_load(f)


def 生成当前木兰codegen的测试结果(source_dict: Dict[str, str], 输出log=False):
    结果 = {}

    for 路径, 源码 in source_dict.items():
        if 输出log:
            print('-----------\n生成中...: %s' % 路径)

        try:
            语法分析器_ = 语法分析器(tokenizer)
            节点 = 语法分析器_.分析(源码, 路径)
            代码生成器 = CodegenNow()
            源码结果 = 代码生成器.得到源码(节点)
            结果[路径] = 源码结果
            if 输出log:
                print(源码结果)
        except 语法错误 as e:
            print(str(e))

    with open('测试/unittest/当前木兰输出结果.json', 'w') as 文件:
        dump(结果, 文件)
    return 结果


def 生成原始木兰的测试结果(原始木兰执行命令: str, 源码字典: Dict[str, str]):
    结果 = {}

    计数 = 0
    源码字典长度 = len(源码字典)

    print('得到原始木兰生成结果中...')

    for 路径, source in 源码字典.items():
        进程 = Popen('%s %s' % (原始木兰执行命令, 路径), stdout=PIPE, stderr=PIPE)
        输出流, 错误输出流 = 进程.communicate()

        原始木兰输出结果 = 输出流.decode()
        
        if 错误输出流.decode('gbk').startswith('SyntaxError') or \
                错误输出流.decode('gbk').startswith('UnicodeDecodeError'):
            continue

        结果[路径] = 原始木兰输出结果.replace('\r\n', '\n')[:-1]  # 除去末尾因为popen读取而导致的 \n

        print('\r进度: （%s/%s）' % (计数, 源码字典长度), end='')

    print()  # 换行
    
    with open('测试/unittest/原始木兰输出结果.json', 'w') as 文件:
        dump(结果, 文件)
    return 结果


class 木兰到Python测试(TestCase):
    def test(self):
        生成测试用例json文件()
        生成当前木兰codegen的测试结果(读取测试用例json())
        生成原始木兰的测试结果(木兰原始文件执行命令, 读取测试用例json())

        with open('测试/unittest/原始木兰输出结果.json', 'r') as f:
            原始木兰输出结果 = json_load(f)
        with open('测试/unittest/当前木兰输出结果.json', 'r') as f:
            当前木兰输出结果 = json_load(f)
        源码字典 = 读取测试用例json()

        for 路径 in 源码字典.keys():
            if 路径 in 原始木兰输出结果 and 路径 in 当前木兰输出结果:
                print('比较 %s' % 路径)
                self.assertEqual(
                    原始木兰输出结果[路径], 当前木兰输出结果[路径], "输出与原始木兰不一致：\n%s%s" % 
                        ('----原始木兰输出结果:\n%s' % repr(原始木兰输出结果[路径]),
                         '----当前木兰输出结果:\n%s' % repr(当前木兰输出结果[路径]),
                        ))



def 插入排序(数列):
  for i in range(1, len(数列)):
    挑出 = 数列[i]
    位置 = i - 1
    while (位置 >= 0 and 数列[位置] > 挑出):
      数列[位置 + 1] = 数列[位置]
      位置 -= 1
    数列[位置 + 1] = 挑出
乱序数列 = [5, 2, 1, 8, 4]
插入排序(乱序数列)
print(乱序数列)


from 木兰.分析器.语法分析器 import 语法分析器


def 生成语法树(源码):
    分析器 = 语法分析器()
    return 分析器.分析(源码, '')

def 读源码生成树(源码文件):
    with open(源码文件, 'r', encoding='utf-8') as f:
        源码 = f.read()
        return 生成语法树(源码)


# 参考: https://stackabuse.com/sorting-algorithms-in-python/
def 冒泡(数列):
  换过 = True
  while 换过:
    换过 = False
    for i in range(0, len(数列) - 1):
      前项 = 数列[i]
      后项 = 数列[i + 1]
      if 前项 > 后项:
        数列[i], 数列[i + 1] = 后项, 前项
        换过 = True

打乱数列 = [5, 2, 1, 8, 4]
冒泡(打乱数列)
print(打乱数列)


class Animal:
  def __init__(self):
    print(1)

  def run(self, n):
    print(n)

class WildAnimal:
  def __init__(self, name):
    print(name)

class Person(Animal):
  def __init__(self):
    super().__init__()

  def go(self):
    super().run(3)

class Wolf(WildAnimal):
  def __init__(self):
    super().__init__(2)

Person()
Wolf()
Person().go()


import codegen
import ast

木兰预置函数映射表 = {
    'println': 'print',
    'tuple': '',
    'char': 'chr',
    'isa': 'isinstance',
    'ceil': 'math.ceil',
    'floor': 'math.floor',
    'fabs': 'math.fabs',
    'sqrt': 'math.sqrt',
    'log': 'math.log',
    'log10': 'math.log10',
    'exp': 'math.exp',
    'pow': 'math.pow',
    'sin': 'math.sin',
    'cos': 'math.cos',
    'tan': 'math.tan',
    'asin': 'math.asin',
    'acos': 'math.acos'
}


class 代码生成器(codegen.SourceGenerator):
    def __init__(self):
        super().__init__('  ', False)

    def visit_arg(self, 参数节点):
        super().write(参数节点.arg)

    def visit_AnnAssign(self, 节点):
        self.newline(节点)
        self.visit(节点.target)
        self.write(' : ')
        self.visit(节点.annotation)
        self.write(' = ')
        self.visit(节点.value)

    def visit_Call(self, 节点):
        if isinstance(节点.func, ast.Name):
            if 节点.func.id in 木兰预置函数映射表:
                节点.func.id = 木兰预置函数映射表[节点.func.id]
        super().visit_Call(节点)

    def visit_ExtSlice(self, 节点):
        for 索引, 切片项 in enumerate(节点.dims):
            if 索引 != 0:
                self.write(', ')
            self.visit(切片项)

    def visit_ImportFrom(self, 节点):
        self.newline(节点)
        if 节点.module:
            self.write('from %s%s import ' % ('.' * 节点.level, 节点.module))
        else:
            self.write('from %s import ' % ('.' * 节点.level))

        for 索引, 项 in enumerate(节点.names):
            if 索引:
                self.write(', ')
            self.visit(项)

    def visit_Name(self, 节点):
        if 节点.id == 'PI':
            self.write('pi')
        else:
            super().visit_Name(节点)

    def visit_NameConstant(self, 节点):
        if 节点.value is None:
            self.write('None')
        elif 节点.value:
            self.write('True')
        else:
            self.write('False')

    def visit_withitem(self, 节点):
        self.visit(节点.context_expr)

        if 节点.optional_vars is not None:
            self.write(' as ')
            self.visit(节点.optional_vars)

    def visit_With(self, 节点):
        self.newline(节点)
        self.write('with ')

        for 索引, 项 in enumerate(节点.items):
            if 索引 > 0:
                self.write(', ')
            self.visit(项)

        self.write(':')
        self.body(节点.body)

    def signature(self, 节点):
        需要逗号 = []

        def 写逗号至结果():
            if 需要逗号:
                self.write(', ')
            else:
                需要逗号.append(True)

        对齐辅助列表 = [None] * (len(节点.args) - len(节点.defaults))

        for 参数, 默认项 in zip(节点.args, 对齐辅助列表 + 节点.defaults):
            写逗号至结果()
            self.visit(参数)
            if 默认项 is not None:
                self.write('=')
                self.visit(默认项)

        if 节点.vararg is not None:
            写逗号至结果()
            self.write('*' + 节点.vararg.arg)

        if 节点.kwarg is not None:
            写逗号至结果()
            self.write('**' + 节点.kwarg.arg)

    def 得到源码(self, 节点):
        self.visit(节点)

        self.result.insert(
            0, 'import sys\nfrom math import *\nARGV = sys.argv[1:]\n')

        return ''.join(self.result)



import sys, cmd; from 随机数 import 随机范围数

class 猜数字(cmd.Cmd):
    intro, prompt = "我想了个 100 之内的数，猜猜是几？", '请猜吧: '
    想的 = 随机范围数(1000) // 10

    def default(self, 行):
        try:
            数 = int(行)
        except ValueError as 例外:
            print(行 + " 不是数，请再试")
            return
        self.比较(数)

    def 比较(self, 数):
        if 数 == self.想的:
            print("中了!")
            sys.exit()
        print("太大了!" if 数 > self.想的 else "太小了!")

猜数字().cmdloop()


from 木兰.生成 import 木兰

import ast
import unittest

from 测试.unittest.生成源码表 import 木兰源码表, python源码表

头部信息 = "/* 本文件由命令 `木兰 -兰 ` 自动生成. */\n"

# 注意：所有测试用源码切勿在末尾加空行，否则比较有误

# TODO: 确保生成的木兰代码运行效果与 Python 相同
# 如果用例与木兰功能测试用例有相当重合，考虑复用


class test所有(unittest.TestCase):

    def test(self):
        for python路径 in 木兰源码表:
            python源码 = python源码表[python路径]
            木兰源码 = 木兰源码表[python路径]

            语法树节点 = ast.parse(python源码, python路径)
            生成器 = 木兰.木兰生成器("  ", 头部信息)
            生成器.visit(语法树节点)
            生成源码 = "".join(生成器.结果)
            self.assertEqual(生成源码, 头部信息 + "\n" + 木兰源码, python路径 + " 转换错误")



import unittest

from 编辑器.运行木兰 import 运行木兰代码
from 木兰.分析器.错误 import 语法错误
from 木兰.功用.反馈信息 import *
from 木兰.共享 import python3版本号
from 测试.unittest.功用 import *

from pathlib import Path

class test所有(unittest.TestCase):

    def test_运行前报错(self):
        对应报错 = {
            "测试/错误处理/2.ul": "没认出这个词 \"整数\"",

            # TODO：如果是用错关键词，如 throw 用错为 raise（原 py 用户）则应提示
            "测试/错误处理/不识关键词.ul": "没认出这个词 \"标识符\"",

            "测试/错误处理/列表内容末尾逗号.ul": "没认出这个词 \"]\"",
            "测试/错误处理/属性为关键词self.ul": "没认出这个词 \"while\"",
            "测试/错误处理/括号不匹配.ul": "没认出这个词 \"$end\"",
            "测试/错误处理/自身.ul": "没认出这个词 \".\"",
            "测试/错误处理/范围by2.ul": "没认出这个词 \"by\"",
            "测试/错误处理/运算换行and.ul": "没认出这个词 \"and\"",
            "测试/错误处理/不可见字符.ul": "分词时没认出这个词 \"\x08\"",
            "测试/错误处理/形参默认值不全.ul": "参数“y”之前的“x”有默认值，它也需要一个表达式指定默认值",
            "测试/错误处理/两参数重载.ul": "双参数仅对 [] 有效，详见《操作符》文档",
            "测试/错误处理/重载减空小括号.ul": "没认出这个词 \")\"",
        }
        for 文件 in 对应报错:
            try:
                读源码生成树(文件)
                self.fail("不该到这")
            except 语法错误 as e:
                self.assertEqual(e.信息, 对应报错[文件], 文件)

    def test_运行时报错(self):
        对应报错 = {
            "测试/错误处理/下标越界.ul": 报错_列表索引,

            # TODO: catch 的类型指定尚无实际作用
            "测试/错误处理/catch类型错误.ul": 报错_列表索引,
            "测试/错误处理/try随意.ul": "需要添加此属性：__enter__" + 参考_enter,
            "测试/错误处理/全局.ul": "请先对本地变量‘x’赋值再引用",
            "测试/错误处理/调用错误函数.ul": "请先定义‘b’再使用",
            "测试/错误处理/多行除零.ul": 报错_除零,
            "测试/错误处理/字典无键.ul": "字典中不存在此键：4",
            "测试/错误处理/字符串拼接.ul": "字符串只能拼接字符串，请将“int”先用 str() 转换",
            "测试/错误处理/属性被静态调用.ul": "函数没有属性‘var’，看看‘self’",
            "测试/错误处理/空取属性.ul": "空变量没有属性‘name’，看看‘a’",
            "测试/错误处理/未定义.ul": "请先定义‘number’再使用",
            "测试/错误处理/模块无属性.ul": "AttributeError：module 'TypeDef' has no attribute 'a'",
            "测试/错误处理/死递归.ul": 报错_递归,
            "测试/错误处理/类型定义中使用本类型.ul": "请先定义‘Person’再使用",
            "测试/错误处理/调用非静态方法.ul": "类型错误：getAge() missing 1 required positional argument: 'self'",
            "测试/错误处理/重复引用_绝对路径1.ul": "AttributeError：module 'test' has no attribute 'TypeDef'",
            "测试/错误处理/重复引用_绝对路径2.ul": "AttributeError：module 'test' has no attribute 'Instance1'",
            "测试/错误处理/空误作数组.ul": "空变量不支持按索引取项，看看‘a’",
            "测试/错误处理/数误作数组.ul": "整数变量不支持按索引取项，看看‘a’",
            "测试/错误处理/函数误作数组.ul": "函数不支持按索引取项，看看‘a’",
            "测试/错误处理/真误作数组.ul": "真假变量不支持按索引取项，看看‘a’",
            "测试/错误处理/真误作数组_部分解析.ul": "真假变量不支持按索引取项",
            "测试/错误处理/无法调用.ul": "类型错误：'int' object is not callable",
            "测试/错误处理/异常十六进制数.ul": "请先定义‘xg’再使用",
        }
        if python3版本号 >= 10:
            对应报错["测试/错误处理/调用非静态方法.ul"] = (
                "类型错误：Person.getAge() missing 1 required positional argument: 'self'"
            )

        for 文件 in 对应报错:
            报错 = 运行木兰代码(文件)
            self.assertEqual(报错[0], 对应报错[文件], 文件)

        单层报错 = {
            "测试/错误处理/catch2.ul": "语法错误：default 'except:' must be last (catch2.ul, line 2)",
            "测试/错误处理/函数外return.ul": "语法错误：'return' outside function (函数外return.ul, line 2)",
        }
        if python3版本号 >= 10:
            单层报错["测试/错误处理/catch2.ul"] = (
                "语法错误：default 'except:' must be last (catch2.ul, line 3)"
            )

        for 文件 in 单层报错:
            报错 = 运行木兰代码(文件)
            self.assertEqual(报错[0], 单层报错[文件], 文件)

        报错 = 运行木兰代码("测试/错误处理/误用函数.ul")
        self.assertEqual(报错[0], "类型错误：'function' object cannot be interpreted as an integer")

        # TODO: 需捕获此错误
        try:
            运行木兰代码("测试/错误处理/非法赋值.ul")
        except ValueError as e:
            self.assertEqual(str(e), "expression which can't be assigned to in Store context")

    def test_行号(self):
        try:
            读源码生成树("测试/错误处理/属性为关键词引用.ul")
            self.fail("不该到这")
        except 语法错误 as e:
            self.assertEqual(e.信息, "没认出这个词 \"while\"")
            self.assertEqual(e.行号, 7)  # 第三行定义可以, 第七行引用时报错

    def test_列号(self):
        try:
            节点 = 生成语法树("using func")
            self.fail("不该到这")
        except 语法错误 as e:
            self.assertEqual(e.列号, 7)

    def test_行列号(self):
        try:
            读源码生成树("测试/错误处理/词不识.ul")
        except 语法错误 as e:
            self.assertEqual(e.信息, "分词时没认出这个词 \"`\"")
            self.assertEqual(e.行号, 3)
            self.assertEqual(e.列号, 1)

    def test_层级(self):
        报错 = 运行木兰代码("测试/错误处理/引用模块.ul")
        self.assertEqual(报错[0], 报错_列表索引)
        self.assertEqual(报错[1], f"“{Path('测试/错误处理/下标越界函数.ul')}”第2行：print([][0])")
        self.assertEqual(报错[2], 报错_层级)
        self.assertEqual(报错[3], "见第3行：a()")

        报错 = 运行木兰代码("测试/错误处理/未定义变量于多层函数.ul")
        self.assertEqual(报错[1], "见第2行：return 数1 + 1")
        self.assertEqual(报错[2], 报错_层级)
        self.assertEqual(报错[3], "见第7行：输出(加(2))")

        报错 = 运行木兰代码("测试/错误处理/引用问题模块.ul")
        self.assertEqual(报错[1], f"“{Path('测试/错误处理/无此变量.ul')}”第1行：a")
        self.assertEqual(报错[2], 报错_层级)
        self.assertEqual(报错[6], "见第1行：using * in 测试.错误处理.无此变量") # 2 6 行之间为木兰源码

        报错 = 运行木兰代码("测试/错误处理/循环引用/a.ul")
        
        self.assertEqual(报错[0], 报错_递归) # 间隔为 ast.py 与木兰源码

        报错内容1 = f"“{Path('测试/错误处理/循环引用/b.ul')}”第1行：using * in 测试.错误处理.循环引用.a"
        报错内容2 = "见第1行：using * in 测试.错误处理.循环引用.b"

        if python3版本号 == 7:
            self.assertEqual(报错[19], 报错内容1)
            self.assertEqual(报错[23], 报错内容2)
        elif 8 <= python3版本号 <= 10:
            self.assertEqual(报错[22], 报错内容1)
            self.assertEqual(报错[26], 报错内容2)
        elif 11 <= python3版本号 <= 12:
            self.assertEqual(报错[23], 报错内容1)
            self.assertEqual(报错[27], 报错内容2)
        else:
            self.fail("请使用 Python 3.7 ~ 3.12 版本运行此测试")



import ast

# 提取所有标识符后分析使用的术语与短语句式

class 标识符提取器(ast.NodeVisitor):

    def __init__(self):
        self.调试 = True
        self.所有标识符 = []

    def 日志(self, 信息):
        if self.调试:
            print(信息)

    def 记录(self, 文本):
        if not isinstance(文本, str):
            raise AssertionError("未正确处理节点：" + repr(文本))
        self.所有标识符.append(文本)

    def visit_Name(self, 节点):
        self.日志("Name: " + 节点.id)
        self.记录(节点.id)

python路径 = '木兰/分析器/词法分析器.py' # '辅助工具/标识符提取.py'
with open(python路径, 'r', encoding='utf-8') as f:
    python源码 = f.read()
语法树节点 = ast.parse(python源码, python路径)
提取器 = 标识符提取器()
提取器.visit(语法树节点)


from ast import *
from 木兰.分析器.语法树 import 语法树

'''
注释"研究"的待进一步揣摩
'''
一元操作符 = {
    Not: '!',
    USub: '-',
}
二元操作符 = {
    Add: '+',
    Sub: '-',
    Mult: '*',
    Div: '/',
    FloorDiv: '/',
    Mod: '%',
    Pow: '^',
}
布尔操作符 = {
    And: 'and',
    Or: 'or',
}
比较操作符 = {
    Eq: '==',
    NotEq: '!=',
    Is: '===',
    IsNot: '!==',
    LtE: "<=",
    Lt: '<',
    In: 'in',
    NotIn: "not in",
    GtE: '>=',
    Gt: '>',
}


def 转源码(节点, 缩进量="  "):
    """
    本方法由语法树生成木兰源码，可用于实现 Python 到木兰源码的简单转换工具。
    """
    自述 = "/* 本文件由命令 `木兰 -兰 ` 自动生成. */\n"
    生成器 = 木兰生成器(缩进量, 头部=自述)
    生成器.visit(节点)
    return "".join(生成器.结果)


# TODO: 将各 visit_ 方法名中文化（方法赋值）
class 木兰生成器(NodeVisitor):

    def __init__(self, 缩进量, 头部=None):
        self.调试 = False
        self.结果 = []
        self.缩进量 = 缩进量
        self.缩进 = 0
        self.行数 = 0
        self.所有类型 = []
        if 头部 is not None:
            self.结果.append(头部)

    def 记录(self, 信息):
        if self.调试:
            print(信息)

    def 编写(self, 文本):
        if not isinstance(文本, str):
            raise AssertionError("未正确处理节点：" + repr(文本))
        if self.行数:
            if self.结果:
                self.结果.append('\n' * self.行数)
            self.结果.append(self.缩进量 * self.缩进)
            self.行数 = 0
        self.结果.append(文本)

    # 待加行号信息
    def 另起一行(self, 节点=None, 额外=0):
        self.行数 = max(self.行数, 1 + 额外)

    def 主体(self, 所有声明):
        self.编写(' {')
        self.缩进 += 1
        for 声明 in 所有声明:
            self.visit(声明)

        self.缩进 -= 1
        self.另起一行()
        self.编写('}')

    def 类型主体(self, 所有声明):
        self.编写(' {')
        self.缩进 += 1
        变量声明 = []

        for 声明 in 所有声明:
            if isinstance(声明, FunctionDef) or isinstance(声明, ClassDef):
                if 变量声明:
                    self.另起一行()
                    self.主体(变量声明)
                    变量声明 = []
                self.visit(声明)
            else:
                变量声明.append(声明)

        if 变量声明:
            self.另起一行()
            self.主体(变量声明)
        self.缩进 -= 1
        self.另起一行()
        self.编写('}')

    def visit_Assign(self, 节点):
        self.另起一行(节点)

        for 索引, 目标 in enumerate(节点.targets):
            if 索引 > 0:
                self.编写(', ')
            self.visit(目标)

        self.编写(' = ')
        self.visit(节点.value)

    def visit_AugAssign(self, 节点):
        self.另起一行(节点)
        self.visit(节点.target)
        self.编写(' ' + 二元操作符[type(节点.op)] + '= ')
        self.visit(节点.value)

    def 形参(self, 节点):
        # TODO: 避免重复
        需逗号 = []

        def 写逗号():
            if 需逗号:
                self.编写(', ')
            else:
                需逗号.append(True)

        空档 = [None] * (len(节点.args) - len(节点.defaults))
        for 形参, 默认值 in zip(节点.args, 空档 + 节点.defaults):
            写逗号()
            self.visit(形参)
            if 默认值 is not None:
                self.编写("=")
                self.visit(默认值)

        # 实际上木兰的变长形参并非用 * 声明，不知此段何用
        '''
        if 节点.vararg is not None:
            写逗号()
            self.编写('*' + 节点.vararg)
        if 节点.kwarg is not None:
            写逗号()
            self.编写("**" + 节点.kwarg)
        '''

    def visit_FunctionDef(self, 节点):
        self.另起一行(额外=1)
        self.另起一行(节点)

        取值 = False
        赋值 = False
        if 节点.decorator_list:

            # 研究：此处会报错，应为 节点.decorator_list
            for 修饰 in decorator_list:
                print(修饰)
                if isinstance(修饰, Name) and 修饰.id == 'property':
                    取值 = True
                elif isinstance(修饰, Attribute) and 修饰.attr == "setter":
                    赋值 = True
        if 取值:
            assert not 赋值
        if len(self.所有类型) > 0:
            if len(节点.args.args) > 0:
                if 节点.args.args[0].arg == 'self':
                    节点.args.args = 节点.args.args[1:]
                    if 节点.name == '__init__':
                        节点.name = self.所有类型[-1]
                    节点.name = '$' + 节点.name
        if 取值 or 赋值:
            self.编写('attr ')
        else:
            self.编写('func ')

        self.编写('%s%s(' % (节点.name, ' = ' if 赋值 else ''))
        self.visit(节点.args)
        self.编写(')')
        self.主体(节点.body)

    def visit_ClassDef(self, 节点):
        self.所有类型.append(节点.name)
        self.另起一行(额外=2)
        self.另起一行(节点)
        self.编写('type %s' % 节点.name)
        for 基类 in 节点.bases:
            if 基类 != 节点.bases[0]:
                self.编写(', ')
            else:
                self.编写(' : ')
            self.visit(基类)

        self.类型主体(节点.body)

        # 研究：为何不 :-1 ？
        self.所有类型 = self.所有类型[:-2]

    def visit_If(self, 节点):
        self.另起一行(节点)
        self.编写('if ')
        self.visit(节点.test)
        self.主体(节点.body)
        while True:
            否则 = 节点.orelse
            if len(否则) == 0:
                break
            elif len(否则) == 1 and isinstance(否则[0], If):
                节点 = 否则[0]
                self.编写(' elif ')
                self.visit(节点.test)
                self.主体(节点.body)
            else:
                self.编写(' else')
                self.主体(否则)
                break

    def visit_For(self, 节点):
        self.另起一行(节点)
        self.编写('for ')
        self.visit(节点.target)
        self.编写(' in ')
        self.visit(节点.iter)
        self.主体(节点.body)

    def visit_While(self, 节点):
        self.另起一行(节点)
        self.编写('while ')
        self.visit(节点.test)
        self.主体(节点.body)

    def visit_Tuple(self, 节点):
        if isinstance(节点.ctx, Load):
            self.编写('tuple(')
        for 索引, 元素 in enumerate(节点.elts):
            if 索引:
                self.编写(', ')
            self.visit(元素)

        if isinstance(节点.ctx, Load):
            self.编写(')')

    def visit_Try(self, 节点):
        self.另起一行(节点)
        self.编写('try')
        self.主体(节点.body)
        for handler in 节点.handlers:
            self.visit(handler)

        # TODO: final

    def visit_Return(self, 节点):
        self.另起一行(节点)
        self.编写('return')
        if 节点.value is not None:
            self.编写(' ')
            self.visit(节点.value)

    def visit_Attribute(self, 节点):
        if isinstance(节点.value, Name) and 节点.value.id == 'self':
            self.编写('$%s' % 节点.attr)
        elif isinstance(节点.value, Call) and isinstance(节点.value.func, Name) and 节点.value.func.id == 'super':
            self.编写('super')
            if 节点.attr != '__init__':
                self.编写('.' + 节点.attr)
        else:
            self.visit(节点.value)
            self.编写('.' + 节点.attr)

    def visit_Call(self, 节点):
        需逗号 = []

        def 写逗号():
            if 需逗号:
                self.编写(', ')
            else:
                需逗号.append(True)

        self.visit(节点.func)
        self.编写('(')
        for 实参 in 节点.args:
            写逗号()
            self.visit(实参)

        for 关键词 in 节点.keywords:
            写逗号()
            self.编写(关键词.arg + '=')
            self.visit(关键词.value)

        self.编写(')')

    def visit_Name(self, 节点):
        self.记录("Name: " + 节点.id)
        if 节点.id == 'print':
            self.编写('println')
        elif 节点.id == 'chr':
            self.编写('char')
        else:
            self.编写(节点.id)

    def visit_Str(self, 节点):
        self.编写(repr(节点.s))

    def visit_Num(self, 节点):
        self.编写(repr(节点.n))

    def 序列(左括号, 右括号):
        def visit(self, 节点):
            self.编写(左括号)
            for 索引, 项 in enumerate(节点.elts):
                if 索引:
                    self.编写(', ')
                self.visit(项)

            self.编写(右括号)

        return visit

    visit_List = 序列('[', ']')

    def visit_Dict(self, 节点):
        self.编写('{')
        为空 = True
        for 序号, (键, 值) in enumerate(zip(节点.keys, 节点.values)):
            为空 = False
            if 序号:
                self.编写(', ')
            self.visit(键)
            self.编写(': ')
            self.visit(值)

        if 为空:
            self.编写(':')
        self.编写('}')

    def visit_BinOp(self, 节点):
        self.visit(节点.left)
        self.编写(' %s ' % 二元操作符[type(节点.op)])
        self.visit(节点.right)

    def visit_BoolOp(self, 节点):
        self.编写('(')
        for 索引, 值 in enumerate(节点.values):
            if 索引:
                self.编写(' %s ' % 布尔操作符[type(节点.op)])
            self.visit(值)

        self.编写(')')

    def visit_Compare(self, 节点):
        self.编写('(')
        左边 = 节点.left
        已开头 = False
        for 操作符, 右边 in zip(节点.ops, 节点.comparators):
            if 已开头:
                self.编写(' and ')
            操作符 = 比较操作符[type(操作符)]
            if 'in' in 操作符:
                if "not" in 操作符:
                    self.编写("!")
                self.visit(右边)
                self.编写('.__contains__(')
                self.visit(左边)
                self.编写(')')
            else:
                self.visit(左边)
                self.编写(' %s ' % 操作符)
                self.visit(右边)
            左边 = 右边
            已开头 = True

        self.编写(')')

    # 在 Python3.8 中 Num, Str, NameConstant 节点
    # 统一被 Constant 节点代替，所以使用一个
    # visit_Constant 方法来统一处理三个情况
    def visit_Constant(self, 节点):
        if 语法树.节点为空(节点):
            self.编写('nil')
        elif 语法树.节点为真假值(节点):
            if literal_eval(节点):
                self.编写('true')
            else:
                self.编写('false')
        elif 语法树.节点为字符串(节点):
            self.编写(repr(literal_eval(节点)))
        elif 语法树.节点为数字(节点):
            self.编写(repr(literal_eval(节点)))
        else:
            assert False, "未知的 Constant 节点类型" + repr(节点)

    def visit_UnaryOp(self, 节点):
        self.编写("(")
        操作符 = 一元操作符[type(节点.op)]
        self.编写(操作符)
        self.visit(节点.operand)
        self.编写(")")

    def visit_Subscript(self, 节点):
        self.visit(节点.value)
        self.编写('[')
        self.visit(节点.slice)
        self.编写(']')

    def visit_Slice(self, 节点):
        if 节点.lower is not None:
            self.visit(节点.lower)
        self.编写(':')
        if 节点.upper is not None:
            self.visit(节点.upper)

    def visit_Import(self, node):
        self.另起一行(node)
        self.编写('using ')
        for 索引, 包 in enumerate(node.names):
            if 索引:
                self.编写(', ')
            self.编写(包.name)

        # TODO: 下两行何用？
        # for 包 in node.names:
        #    self.visit(包)

    def visit_ImportFrom(self, node):
        self.另起一行(node)
        self.编写('using ')
        for 索引, 包 in enumerate(node.names):
            if 索引:
                self.编写(', ')
            self.编写(包.name)

        if node.module:
            self.编写(' in %s%s' % ('.' * node.level, node.module))
        else:
            self.编写(' in %s' % ('.' * node.level))

        # TODO: 下两行何用？
        # for item in node.names:
        #    self.visit(item)

    def visit_IfExp(self, 节点):
        self.visit(节点.test)
        self.编写(' ? ')
        self.visit(节点.body)
        self.编写(' : ')
        self.visit(节点.orelse)

    def visit_Expr(self, 节点):
        self.记录("Expr: " + str(节点))
        self.另起一行()
        节点值 = 节点.value
        if 语法树.节点为字符串(节点值):
            self.编写('/* %s */' % literal_eval(节点值))
        else:
            self.visit(节点值)

    def visit_FormattedValue(self, 节点):
        self.编写('str(')
        self.visit(节点.value)
        self.编写(')')

    def visit_NameConstant(self, 节点):
        if 节点.value is None:
            self.编写('nil')
        elif 节点.value:
            self.编写('true')
        else:
            self.编写('false')

    def visit_ExceptHandler(self, 节点):
        self.另起一行(节点)
        self.编写('catch ')
        if 节点.name is not None:
            # 此处原为 self.visit(节点.name)，会报错 AttributeError: 'str' object has no attribute '_fields'
            self.编写(节点.name)
        else:
            self.编写('__')
        if 节点.type is not None:
            self.编写(' : ')
            self.visit(节点.type)
        self.主体(节点.body)

    def visit_arguments(self, 节点):
        self.形参(节点)

    def visit_arg(self, 节点):
        self.编写(节点.arg)


def 转换(节点):
    return 转源码(节点)



import unittest
import ast

from 木兰.共享 import python3版本号
from 木兰.分析器.词法分析器 import 分词器
from 测试.unittest.功用 import *


# TODO：需确保无此类 Warning：ParserGeneratorWarning: 28 shift/reduce conflicts
class test语法树(unittest.TestCase):

    def test_词法分析(self):
        # 由于换行也包括在内，"后括号"一词的位置并非 2：1
        各词 = self.分词("{print(2)\n}")
        for 词 in 各词:
            最末 = 词
        self.assertEqual(最末.getstr(), "\n}")
        self.assertEqual(最末.getsourcepos().lineno, 1)
        self.assertEqual(最末.getsourcepos().colno, 10)

    def test_行列号(self):
        节点 = 生成语法树("print(1/0)")
        expr节点 = self.取子节点(节点, "body", 0)
        call节点 = self.取子节点(expr节点, "value")
        self.assertEqual(call节点.lineno, 1)
        self.assertEqual(call节点.col_offset, 1)

        除法节点 = self.取子节点(call节点, "func")
        self.assertEqual(除法节点.lineno, 1)
        self.assertEqual(除法节点.col_offset, 1)

    def test_整树比较(self):
        期望值_3_7 = {
            "反斜杠.ul": r"Module(body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=1, col_offset=1), args=[Str(s='\\', lineno=1, col_offset=7)], keywords=[], lineno=1, col_offset=1), lineno=1, col_offset=1)])",
            "带类型参数.ul": r"Module(body=[FunctionDef(name='hello', args=arguments(args=[arg(arg='p', annotation=Name(id='Person', ctx=Load(), lineno=1, col_offset=16), lineno=1, col_offset=12)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Num(n=1, lineno=2, col_offset=10), lineno=2, col_offset=3)], decorator_list=[], lineno=1, col_offset=1)])",
            "类型.ul": r"Module(body=[ClassDef(name='Person', bases=[], keywords=[], body=[FunctionDef(name='__add__', args=arguments(args=[arg(arg='self', lineno=2, col_offset=12), arg(arg='a', annotation=None, lineno=2, col_offset=13)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Call(func=Name(id='Person', ctx=Load(), lineno=2, col_offset=25), args=[BinOp(left=Attribute(value=Name(id='self', ctx=Load(), lineno=2, col_offset=32), attr='name', ctx=Load(), lineno=2, col_offset=32), op=Add(), right=Attribute(value=Name(id='a', ctx=Load(), lineno=2, col_offset=44), attr='name', ctx=Load(), lineno=2, col_offset=44), lineno=2, col_offset=32)], keywords=[], lineno=2, col_offset=25), lineno=2, col_offset=18)], decorator_list=[], lineno=2, col_offset=1), FunctionDef(name='__init__', args=arguments(args=[arg(arg='self', annotation=None, lineno=3, col_offset=1), arg(arg='name', annotation=None, lineno=3, col_offset=14)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Assign(targets=[Attribute(value=Name(id='self', ctx=Load(), lineno=3, col_offset=22), attr='name', ctx=Store(), lineno=3, col_offset=22)], value=Name(id='name', ctx=Load(), lineno=3, col_offset=34), lineno=3, col_offset=22)], decorator_list=[], lineno=3, col_offset=1)], decorator_list=[], lineno=1, col_offset=1)])",
            "函数.ul": r"Module(body=[FunctionDef(name='echo', args=arguments(args=[arg(arg='number', annotation=None, lineno=1, col_offset=11)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Name(id='number', ctx=Load(), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], decorator_list=[], lineno=1, col_offset=1), Expr(value=Call(func=Name(id='echo', ctx=Load(), lineno=4, col_offset=1), args=[Num(n=2, lineno=4, col_offset=6)], keywords=[], lineno=4, col_offset=1), lineno=4, col_offset=1)])",
            "空匿名函数.ul": r"Module(body=[Assign(targets=[Name(id='b', ctx=Store(), lineno=1, col_offset=1)], value=Lambda(args=arguments(args=[], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=Num(n=2, lineno=1, col_offset=11), lineno=0, col_offset=0), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Call(func=Name(id='b', ctx=Load(), lineno=2, col_offset=7), args=[], keywords=[], lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)])",
            "赋值.ul": r"Module(body=[Assign(targets=[Name(id='a', ctx=Store(), lineno=1, col_offset=1)], value=Num(n=2, lineno=1, col_offset=3), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Name(id='a', ctx=Load(), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)])",
            "多行运算.ul": r"Module(body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=1, col_offset=1), args=[Num(n=0, lineno=1, col_offset=7)], keywords=[], lineno=1, col_offset=1), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[BinOp(left=BinOp(left=Num(n=1, lineno=2, col_offset=7), op=Add(), right=BinOp(left=Num(n=2, lineno=2, col_offset=9), op=Mult(), right=Num(n=3, lineno=2, col_offset=11), lineno=2, col_offset=9), lineno=2, col_offset=7), op=Sub(), right=Call(func=Name(id='__div__', ctx=Load(), lineno=2, col_offset=13), args=[Num(n=4, lineno=2, col_offset=13), Num(n=5, lineno=2, col_offset=15)], keywords=[], lineno=2, col_offset=13), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)])",
            "超类语法.ul": r"Module(body=[ClassDef(name='Person', bases=[Name(id='list', ctx=Load(), lineno=1, col_offset=15)], keywords=[], body=[FunctionDef(name='__init__', args=arguments(args=[arg(arg='self', annotation=None, lineno=2, col_offset=3)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Expr(value=Call(func=Attribute(value=Call(func=Name(id='super', ctx=Load(), lineno=3, col_offset=5), args=[], keywords=[], lineno=3, col_offset=5), attr='__init__', ctx=Load(), lineno=3, col_offset=5), args=[], keywords=[], lineno=3, col_offset=5), lineno=3, col_offset=5)], decorator_list=[], lineno=2, col_offset=3)], decorator_list=[], lineno=1, col_offset=1), Assign(targets=[Name(id='p', ctx=Store(), lineno=6, col_offset=1)], value=Call(func=Name(id='Person', ctx=Load(), lineno=6, col_offset=5), args=[], keywords=[], lineno=6, col_offset=5), lineno=6, col_offset=1), Assign(targets=[Name(id='v', ctx=Store(), lineno=7, col_offset=1)], value=Call(func=Attribute(value=Name(id='p', ctx=Load(), lineno=7, col_offset=5), attr='super', ctx=Load(), lineno=7, col_offset=5), args=[], keywords=[], lineno=7, col_offset=7), lineno=7, col_offset=1), Expr(value=Call(func=Attribute(value=Name(id='Person', ctx=Load(), lineno=8, col_offset=1), attr='super', ctx=Load(), lineno=8, col_offset=1), args=[Name(id='None', ctx=Load(), lineno=8, col_offset=14)], keywords=[], lineno=8, col_offset=8), lineno=8, col_offset=8)])",
            "指定函数类型.ul": r"Module(body=[FunctionDef(name='hello', args=arguments(args=[arg(arg='p', annotation=Name(id='int', ctx=Load(), lineno=1, col_offset=16), lineno=1, col_offset=12)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Name(id='p', ctx=Load(), lineno=2, col_offset=10), lineno=2, col_offset=3)], decorator_list=[], returns=Name(id='int', ctx=Load(), lineno=1, col_offset=23), lineno=1, col_offset=1)])",
            "顺便处理各表达式前缀.ul": r"Module(body=[With(items=[withitem(context_expr=Name(id='c', ctx=Load(), lineno=1, col_offset=9), optional_vars=Tuple(elts=[Name(id='a', ctx=Load(), lineno=1, col_offset=5), Name(id='b', ctx=Load(), lineno=1, col_offset=7)], ctx=Store(), lineno=1, col_offset=5))], body=[Pass( lineno=1, col_offset=11)], lineno=1, col_offset=1)])",
            "善后.ul": r"Module(body=[Try(body=[Pass( lineno=1, col_offset=5)], handlers=[Pass( lineno=1, col_offset=5)], orelse=[], finalbody=[Pass( lineno=1, col_offset=16)], lineno=1, col_offset=1)])",
            "关键字参数.ul": r"Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load(), lineno=1, col_offset=1), args=[], keywords=[keyword(arg='x', value=Num(n=1, lineno=1, col_offset=5))], lineno=1, col_offset=1), lineno=1, col_offset=1)])",
        }

        期望值_3_8 = {
            "反斜杠.ul": r"Module(body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=1, col_offset=1), args=[Constant(value='\\', lineno=1, col_offset=7)], keywords=[], lineno=1, col_offset=1), lineno=1, col_offset=1)], type_ignores=[])",
            "带类型参数.ul": r"Module(body=[FunctionDef(name='hello', args=arguments(posonlyargs=[], args=[arg(arg='p', annotation=Name(id='Person', ctx=Load(), lineno=1, col_offset=16), lineno=1, col_offset=12)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Constant(value=1, lineno=2, col_offset=10), lineno=2, col_offset=3)], decorator_list=[], lineno=1, col_offset=1)], type_ignores=[])",
            "类型.ul": r"Module(body=[ClassDef(name='Person', bases=[], keywords=[], body=[FunctionDef(name='__add__', args=arguments(posonlyargs=[], args=[arg(arg='self', lineno=2, col_offset=12), arg(arg='a', annotation=None, lineno=2, col_offset=13)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Call(func=Name(id='Person', ctx=Load(), lineno=2, col_offset=25), args=[BinOp(left=Attribute(value=Name(id='self', ctx=Load(), lineno=2, col_offset=32), attr='name', ctx=Load(), lineno=2, col_offset=32), op=Add(), right=Attribute(value=Name(id='a', ctx=Load(), lineno=2, col_offset=44), attr='name', ctx=Load(), lineno=2, col_offset=44), lineno=2, col_offset=32)], keywords=[], lineno=2, col_offset=25), lineno=2, col_offset=18)], decorator_list=[], lineno=2, col_offset=1), FunctionDef(name='__init__', args=arguments(posonlyargs=[], args=[arg(arg='self', annotation=None, lineno=3, col_offset=1), arg(arg='name', annotation=None, lineno=3, col_offset=14)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Assign(targets=[Attribute(value=Name(id='self', ctx=Load(), lineno=3, col_offset=22), attr='name', ctx=Store(), lineno=3, col_offset=22)], value=Name(id='name', ctx=Load(), lineno=3, col_offset=34), lineno=3, col_offset=22)], decorator_list=[], lineno=3, col_offset=1)], decorator_list=[], lineno=1, col_offset=1)], type_ignores=[])",
            "函数.ul": r"Module(body=[FunctionDef(name='echo', args=arguments(posonlyargs=[], args=[arg(arg='number', annotation=None, lineno=1, col_offset=11)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Name(id='number', ctx=Load(), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], decorator_list=[], lineno=1, col_offset=1), Expr(value=Call(func=Name(id='echo', ctx=Load(), lineno=4, col_offset=1), args=[Constant(value=2, lineno=4, col_offset=6)], keywords=[], lineno=4, col_offset=1), lineno=4, col_offset=1)], type_ignores=[])",
            "空匿名函数.ul": r"Module(body=[Assign(targets=[Name(id='b', ctx=Store(), lineno=1, col_offset=1)], value=Lambda(args=arguments(posonlyargs=[], args=[], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=Constant(value=2, lineno=1, col_offset=11), lineno=0, col_offset=0), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Call(func=Name(id='b', ctx=Load(), lineno=2, col_offset=7), args=[], keywords=[], lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], type_ignores=[])",
            "赋值.ul": r"Module(body=[Assign(targets=[Name(id='a', ctx=Store(), lineno=1, col_offset=1)], value=Constant(value=2, lineno=1, col_offset=3), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Name(id='a', ctx=Load(), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], type_ignores=[])",
            "多行运算.ul": r"Module(body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=1, col_offset=1), args=[Constant(value=0, lineno=1, col_offset=7)], keywords=[], lineno=1, col_offset=1), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[BinOp(left=BinOp(left=Constant(value=1, lineno=2, col_offset=7), op=Add(), right=BinOp(left=Constant(value=2, lineno=2, col_offset=9), op=Mult(), right=Constant(value=3, lineno=2, col_offset=11), lineno=2, col_offset=9), lineno=2, col_offset=7), op=Sub(), right=Call(func=Name(id='__div__', ctx=Load(), lineno=2, col_offset=13), args=[Constant(value=4, lineno=2, col_offset=13), Constant(value=5, lineno=2, col_offset=15)], keywords=[], lineno=2, col_offset=13), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], type_ignores=[])",
            "超类语法.ul": r"Module(body=[ClassDef(name='Person', bases=[Name(id='list', ctx=Load(), lineno=1, col_offset=15)], keywords=[], body=[FunctionDef(name='__init__', args=arguments(posonlyargs=[], args=[arg(arg='self', annotation=None, lineno=2, col_offset=3)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Expr(value=Call(func=Attribute(value=Call(func=Name(id='super', ctx=Load(), lineno=3, col_offset=5), args=[], keywords=[], lineno=3, col_offset=5), attr='__init__', ctx=Load(), lineno=3, col_offset=5), args=[], keywords=[], lineno=3, col_offset=5), lineno=3, col_offset=5)], decorator_list=[], lineno=2, col_offset=3)], decorator_list=[], lineno=1, col_offset=1), Assign(targets=[Name(id='p', ctx=Store(), lineno=6, col_offset=1)], value=Call(func=Name(id='Person', ctx=Load(), lineno=6, col_offset=5), args=[], keywords=[], lineno=6, col_offset=5), lineno=6, col_offset=1), Assign(targets=[Name(id='v', ctx=Store(), lineno=7, col_offset=1)], value=Call(func=Attribute(value=Name(id='p', ctx=Load(), lineno=7, col_offset=5), attr='super', ctx=Load(), lineno=7, col_offset=5), args=[], keywords=[], lineno=7, col_offset=7), lineno=7, col_offset=1), Expr(value=Call(func=Attribute(value=Name(id='Person', ctx=Load(), lineno=8, col_offset=1), attr='super', ctx=Load(), lineno=8, col_offset=1), args=[Name(id='None', ctx=Load(), lineno=8, col_offset=14)], keywords=[], lineno=8, col_offset=8), lineno=8, col_offset=8)], type_ignores=[])",
            "指定函数类型.ul": r"Module(body=[FunctionDef(name='hello', args=arguments(posonlyargs=[], args=[arg(arg='p', annotation=Name(id='int', ctx=Load(), lineno=1, col_offset=16), lineno=1, col_offset=12)], vararg=None, kwonlyargs=[], kw_defaults=[], kwarg=None, defaults=[]), body=[Return(value=Name(id='p', ctx=Load(), lineno=2, col_offset=10), lineno=2, col_offset=3)], decorator_list=[], returns=Name(id='int', ctx=Load(), lineno=1, col_offset=23), lineno=1, col_offset=1)], type_ignores=[])",
            "顺便处理各表达式前缀.ul": r"Module(body=[With(items=[withitem(context_expr=Name(id='c', ctx=Load(), lineno=1, col_offset=9), optional_vars=Tuple(elts=[Name(id='a', ctx=Load(), lineno=1, col_offset=5), Name(id='b', ctx=Load(), lineno=1, col_offset=7)], ctx=Store(), lineno=1, col_offset=5))], body=[Pass(lineno=1, col_offset=11)], lineno=1, col_offset=1)], type_ignores=[])",
            "善后.ul": r"Module(body=[Try(body=[Pass(lineno=1, col_offset=5)], handlers=[Pass(lineno=1, col_offset=5)], orelse=[], finalbody=[Pass(lineno=1, col_offset=16)], lineno=1, col_offset=1)], type_ignores=[])",
            "关键字参数.ul": r"Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load(), lineno=1, col_offset=1), args=[], keywords=[keyword(arg='x', value=Constant(value=1, lineno=1, col_offset=5))], lineno=1, col_offset=1), lineno=1, col_offset=1)], type_ignores=[])",
        }

        期望值_3_9 = 期望值_3_8.copy()
        期望值_3_9.update(
            {
                "带类型参数.ul": r"Module(body=[FunctionDef(name='hello', args=arguments(posonlyargs=[], args=[arg(arg='p', annotation=Name(id='Person', ctx=Load(), lineno=1, col_offset=16), lineno=1, col_offset=12)], kwonlyargs=[], kw_defaults=[], defaults=[]), body=[Return(value=Constant(value=1, lineno=2, col_offset=10), lineno=2, col_offset=3)], decorator_list=[], lineno=1, col_offset=1)], type_ignores=[])",
                "类型.ul": r"Module(body=[ClassDef(name='Person', bases=[], keywords=[], body=[FunctionDef(name='__add__', args=arguments(posonlyargs=[], args=[arg(arg='self', lineno=2, col_offset=12), arg(arg='a', lineno=2, col_offset=13)], kwonlyargs=[], kw_defaults=[], defaults=[]), body=[Return(value=Call(func=Name(id='Person', ctx=Load(), lineno=2, col_offset=25), args=[BinOp(left=Attribute(value=Name(id='self', ctx=Load(), lineno=2, col_offset=32), attr='name', ctx=Load(), lineno=2, col_offset=32), op=Add(), right=Attribute(value=Name(id='a', ctx=Load(), lineno=2, col_offset=44), attr='name', ctx=Load(), lineno=2, col_offset=44), lineno=2, col_offset=32)], keywords=[], lineno=2, col_offset=25), lineno=2, col_offset=18)], decorator_list=[], lineno=2, col_offset=1), FunctionDef(name='__init__', args=arguments(posonlyargs=[], args=[arg(arg='self', lineno=3, col_offset=1), arg(arg='name', lineno=3, col_offset=14)], kwonlyargs=[], kw_defaults=[], defaults=[]), body=[Assign(targets=[Attribute(value=Name(id='self', ctx=Load(), lineno=3, col_offset=22), attr='name', ctx=Store(), lineno=3, col_offset=22)], value=Name(id='name', ctx=Load(), lineno=3, col_offset=34), lineno=3, col_offset=22)], decorator_list=[], lineno=3, col_offset=1)], decorator_list=[], lineno=1, col_offset=1)], type_ignores=[])",
                "函数.ul": r"Module(body=[FunctionDef(name='echo', args=arguments(posonlyargs=[], args=[arg(arg='number', lineno=1, col_offset=11)], kwonlyargs=[], kw_defaults=[], defaults=[]), body=[Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Name(id='number', ctx=Load(), lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], decorator_list=[], lineno=1, col_offset=1), Expr(value=Call(func=Name(id='echo', ctx=Load(), lineno=4, col_offset=1), args=[Constant(value=2, lineno=4, col_offset=6)], keywords=[], lineno=4, col_offset=1), lineno=4, col_offset=1)], type_ignores=[])",
                "空匿名函数.ul": r"Module(body=[Assign(targets=[Name(id='b', ctx=Store(), lineno=1, col_offset=1)], value=Lambda(args=arguments(posonlyargs=[], args=[], kwonlyargs=[], kw_defaults=[], defaults=[]), body=Constant(value=2, lineno=1, col_offset=11), lineno=0, col_offset=0), lineno=1, col_offset=1), Expr(value=Call(func=Name(id='print', ctx=Load(), lineno=2, col_offset=1), args=[Call(func=Name(id='b', ctx=Load(), lineno=2, col_offset=7), args=[], keywords=[], lineno=2, col_offset=7)], keywords=[], lineno=2, col_offset=1), lineno=2, col_offset=1)], type_ignores=[])",
                "超类语法.ul": r"Module(body=[ClassDef(name='Person', bases=[Name(id='list', ctx=Load(), lineno=1, col_offset=15)], keywords=[], body=[FunctionDef(name='__init__', args=arguments(posonlyargs=[], args=[arg(arg='self', lineno=2, col_offset=3)], kwonlyargs=[], kw_defaults=[], defaults=[]), body=[Expr(value=Call(func=Attribute(value=Call(func=Name(id='super', ctx=Load(), lineno=3, col_offset=5), args=[], keywords=[], lineno=3, col_offset=5), attr='__init__', ctx=Load(), lineno=3, col_offset=5), args=[], keywords=[], lineno=3, col_offset=5), lineno=3, col_offset=5)], decorator_list=[], lineno=2, col_offset=3)], decorator_list=[], lineno=1, col_offset=1), Assign(targets=[Name(id='p', ctx=Store(), lineno=6, col_offset=1)], value=Call(func=Name(id='Person', ctx=Load(), lineno=6, col_offset=5), args=[], keywords=[], lineno=6, col_offset=5), lineno=6, col_offset=1), Assign(targets=[Name(id='v', ctx=Store(), lineno=7, col_offset=1)], value=Call(func=Attribute(value=Name(id='p', ctx=Load(), lineno=7, col_offset=5), attr='super', ctx=Load(), lineno=7, col_offset=5), args=[], keywords=[], lineno=7, col_offset=7), lineno=7, col_offset=1), Expr(value=Call(func=Attribute(value=Name(id='Person', ctx=Load(), lineno=8, col_offset=1), attr='super', ctx=Load(), lineno=8, col_offset=1), args=[Name(id='None', ctx=Load(), lineno=8, col_offset=14)], keywords=[], lineno=8, col_offset=8), lineno=8, col_offset=8)], type_ignores=[])",
                "指定函数类型.ul": r"Module(body=[FunctionDef(name='hello', args=arguments(posonlyargs=[], args=[arg(arg='p', annotation=Name(id='int', ctx=Load(), lineno=1, col_offset=16), lineno=1, col_offset=12)], kwonlyargs=[], kw_defaults=[], defaults=[]), body=[Return(value=Name(id='p', ctx=Load(), lineno=2, col_offset=10), lineno=2, col_offset=3)], decorator_list=[], returns=Name(id='int', ctx=Load(), lineno=1, col_offset=23), lineno=1, col_offset=1)], type_ignores=[])",
                "关键字参数.ul": r"Module(body=[Expr(value=Call(func=Name(id='f', ctx=Load(), lineno=1, col_offset=1), args=[], keywords=[keyword(arg='x', value=Constant(value=1, lineno=1, col_offset=5), lineno=1, col_offset=1)], lineno=1, col_offset=1), lineno=1, col_offset=1)], type_ignores=[])",
            }
        )

        if python3版本号 == 7:
            self._test_整树比较(期望值_3_7)
        elif python3版本号 == 8:
            self._test_整树比较(期望值_3_8)
        elif 9 <= python3版本号 <= 12:
            self._test_整树比较(期望值_3_9)
        else:
            self.fail("请使用 Python 3.7 ~ 3.12 版本运行此测试")

    def _test_整树比较(self, 期望值):
        路径 = "测试/unittest/例程/"
        for 文件 in 期望值:
            源码路径 = 路径 + 文件
            节点 = 读源码生成树(源码路径)
            # 已知3.7.4版本python中ast.Pass生成的语法树多一个空格，高版本没有，下方代码清除测试用例产生的空格与木兰生成的对照中的这
            # 个多余空格，保证在高版本中也可以测试通过。
            self.assertEqual(
                ast.dump(节点, True, True).replace('Pass( lineno', 'Pass(lineno'),
                期望值[文件].replace('Pass( lineno', 'Pass(lineno'),
                f"\"{文件}\"出错"
            )

    def 分词(self, 源码):
        return 分词器.分词(源码)

    def 取子节点(self, 节点, 子节点名, 索引 = -1):
        for 子节点 in ast.iter_fields(节点):
            if 子节点[0] == 子节点名:
                if isinstance(子节点[1], list):
                    return 子节点[1][索引]
                else:
                    return 子节点[1]

if __name__ == '__main__':
    unittest.main()



class Person:
  def __init__(self):
    self.class = 1

mulan = Person()
print(mulan.class)
''' 对比木兰中的 type 属性定义
  File "测试/类型/个体属性class.py", line 3
    self.class = 1
             ^
SyntaxError: invalid syntax
'''


import ast, sys
from io import StringIO
from pathlib import Path

from 木兰.分析器.词法分析器 import 分词器
from 木兰.分析器.语法分析器 import 语法分析器
from 木兰.环境 import 创建全局变量
from 木兰.功用.反馈信息 import 中文化

# 木兰不支持 compile, exec, 因此不改写为木兰代码
def 运行木兰代码(源码文件):
    源码文件 = str(Path(源码文件))
    with open(源码文件, 'r', encoding='utf-8') as f:
        源码 = f.read()

    分析器 = 语法分析器(分词器)
    节点 = 分析器.分析(源码, 源码文件)

    # 参考： https://stackoverflow.com/questions/3906232/python-get-the-print-output-in-an-exec-statement
    原标准输出 = sys.stdout

    # 默认换行符为 \n，见 https://docs.python.org/zh-cn/3.7/library/io.html#io.StringIO
    重定向输出 = sys.stdout = StringIO()

    # 参考：https://docs.python.org/3.7/library/functions.html?highlight=compile#compile
    try:
        可执行码 = compile(节点, 源码文件, 'exec')

        环境变量 = 创建全局变量(文件名=源码文件)

        exec(可执行码, 环境变量)
    except Exception as e:
        return 中文化(e, 源码文件)
    finally:
        sys.stdout = 原标准输出

    return 重定向输出.getvalue()



from pathlib import Path

源码目录 = Path("测试/unittest/源码生成/")

原始木兰未过 = {
    Path("测试/unittest/源码生成/函数/变长指名参数.py"): "TypeError: can only concatenate str (not \"arg\") to str",
    Path("测试/unittest/源码生成/类型/属性.py"): "NameError: name 'decorator_list' is not defined",
    Path("测试/unittest/源码生成/流程控制/try.py"): "AttributeError: 'str' object has no attribute '_fields'",
    Path("测试/unittest/源码生成/算法/排序/冒泡.py"): "UnicodeDecodeError: 'gbk' codec can't decode byte 0x80 in position 6: illegal multibyte sequence",
    Path("测试/unittest/源码生成/综合/猜数字.py"):  "UnicodeDecodeError: 'gbk' codec can't decode byte 0xb0 in position 30: illegal multibyte sequence",
}

python源码表 = {}
木兰源码表 = {}

for 完整路径 in list(源码目录.glob('**/*.py')):
    if 完整路径 not in 原始木兰未过:
        python文件名 = str(完整路径)[:-len(完整路径.suffix)]
        python路径 = python文件名 + '.py'
        with open(python路径, 'r', encoding='utf-8') as f:
            python源码 = f.read()

        木兰路径 = python文件名 + ".ul"
        if Path(木兰路径).exists():
            with open(木兰路径, 'r', encoding='utf-8') as f:
                木兰源码 = f.read()
        else:
            木兰源码 = python源码

        python源码表[python路径] = python源码
        木兰源码表[python路径] = 木兰源码


#!/usr/bin/env python3

import ast
import getopt
import os
import sys

from sys import argv

from 木兰.交互 import 开始交互
from 木兰.分析器.词法分析器 import 分词器
from 木兰.分析器.语法分析器 import 语法分析器
from 木兰.功用.反馈信息 import 反馈信息
from 木兰.环境 import 创建全局变量
from 木兰.生成 import 木兰, python


def 用途(程序):
    介绍 = '''使用方法: %s [-树p码兰调交反执溯版助] 源码文件
    
选项：
 --语法树　　　　　 -树　　将木兰源码转换为 Python 语法树
 --木兰变python　　 -p 　　将木兰源码转换为 Python 源码
 --生成字节码　　　 -码　　将木兰源码转换为 donsok 字节码 (实验性)
 --python变木兰　　 -兰　　将 Python 源码转换为木兰源码
 --调试　　　　　　 -调　　使用 Pdb 环境调试代码
 --交互　　　　　　 -交　　以交互式审查脚本
 --反汇编　　　　　 -反　　将生成的 Python 字节码反汇编
 --执行代码=<代码>　-执　　执行来自命令行参数的代码
 --显示回溯　　　　 -溯　　显示异常的栈回溯信息
 --版本　　　　　　 -版　　显示版本
 --帮助　　　　　　 -助　　显示帮助信息
'''
    sys.stderr.write(介绍 % os.path.basename(程序))
    sys.exit(-1)


def 中(命令行各分段=None):
    if 命令行各分段 is None:
        命令行各分段 = argv

    try:
        所有选项, 参数 = getopt.getopt(
            命令行各分段[1:],
            '版助反交兰p码调溯树执词',
            [
                "语法树",
                "python变木兰",
                '版本',
                '木兰变python',
                '生成字节码',
                '调试',
                '反汇编',
                '执行代码=',
                '显示回溯',
                '帮助',
                '交互',
                '显示分词',
            ]
        )
    except getopt.GetoptError as e:
        try:
            sys.stderr.write(str(e) + '\n')
            用途(命令行各分段[0])
        finally:
            e = None
            del e

    源码文件 = None
    版本 = False
    python变木兰 = False
    语法树 = False
    生成python代码 = False
    生成字节码 = False
    调试 = False
    反汇编 = False
    命令行执行码 = None
    显示回溯 = False
    交互 = False
    显示分词 = False
    for 选项, 值 in 所有选项:
        if 选项 in ('-版', '--版本'):
            版本 = True
        elif 选项 in ("-兰", "--python变木兰"):
            python变木兰 = True
        elif 选项 in ("-树", "--语法树"):
            语法树 = True
        elif 选项 in ('--木兰变python', '-p'):
            生成python代码 = True
        elif 选项 in ('--生成字节码', '-码'):
            生成字节码 = True
        elif 选项 in ('--调试', '-调'):
            调试 = True
        elif 选项 in ('--反汇编', '-反'):
            反汇编 = True
        elif 选项 in ('--显示回溯', '-溯'):
            显示回溯 = True
        elif 选项 in ('--帮助', '-助'):
            用途(命令行各分段[0])
        elif 选项 in ('--交互', '-交'):
            交互 = True
        elif 选项 in ('--执行代码', '-执'):
            命令行执行码 = 值
        elif 选项 in ('--显示分词', '-词'):
            显示分词 = True

    if 版本:
        from 木兰 import __版本__
        sys.stderr.write('%s\n' % __版本__)
        sys.exit()

    if len(命令行各分段) == 1:
        sys.exit(开始交互())

    if 源码文件 is None:
        if len(参数) > 0:
            源码文件 = 参数[0]
        else:
            if not 命令行执行码:
                用途(命令行各分段[0])
    try:
        源码 = None

        if 命令行执行码:
            源码 = 命令行执行码
            源码文件 = '<命令行>'
        elif 源码文件 == '-':
            源码 = sys.stdin.read()
            源码文件 = '<标准输入流>'
        else:
            with open(源码文件, encoding='UTF-8') as 源码文件对象:
                源码 = 源码文件对象.read()

        if not 源码:
            sys.stderr.write('文件 %s 为空！' % 源码文件)
            sys.exit(-1)

        if 显示分词:
            各词 = 分词器.分词(源码)
            for 词 in 各词:
                print(词.gettokentype(), end=' ')
            return

        if python变木兰:
            语法树节点 = ast.parse(源码, 源码文件)
            print(木兰.转换(语法树节点))
            return

        分析器 = 语法分析器(分词器)
        节点 = 分析器.分析(源码, 源码文件)

        if 生成python代码:
            print(python.代码生成器().得到源码(节点))
            return

        if 语法树:
            print(ast.dump(节点, True, True))
            # print(语法树相关.格式化节点(节点, 1))
            return

        if 生成字节码:
            try:
                from pygen.compiler import Compiler
                print(Compiler().compile(节点, 源码文件).dump())
            except ModuleNotFoundError as _:
                sys.stderr.write(f"依赖库 pygen 未找到")
            return

        # 参考：https://docs.python.org/3.7/library/functions.html?highlight=compile#compile
        可执行码 = compile(节点, 源码文件, 'exec')

        if 反汇编:
            from dis import dis
            dis(可执行码)
            return

        环境变量 = 创建全局变量(文件名=源码文件)

        if 调试:
            from pdb import run as Pdb运行, Restart as Pdb重新运行
            while True:
                try:
                    Pdb运行(可执行码, 环境变量, None)
                except Pdb重新运行:
                    pass
                else:
                    break

        else:
            exec(可执行码, 环境变量)

        if 交互:
            sys.exit(开始交互(全局变量=环境变量))

    except Exception as e:
        if 显示回溯:
            raise
        else:
            sys.stderr.write('%s\n' % 反馈信息(e, 源码文件))




import ast
import os
import unittest
import subprocess
from sys import platform

from 测试.unittest.Python到木兰 import *

from 测试.unittest.生成源码表 import 木兰源码表

原始可执行文件 = "ulang-0.2.2.exe"
头部信息 = "/* This file is auto-generated by the command `ulang -s `. */\n"

class test所有(unittest.TestCase):

    def test_python到木兰(self):
        for python路径 in 木兰源码表:
            木兰源码 = 木兰源码表[python路径]

            进程 = subprocess.Popen([原始可执行文件, '-s', python路径], stdout=subprocess.PIPE)
            反馈 = 进程.communicate()
            生成源码 = 反馈[0]
            self.assertEqual(生成源码.decode("utf-8").replace("\r\n", "\n"),
                头部信息 + "\n" + 木兰源码 + "\n",
                python路径 + " 转换错误")